#!/usr/bin/env python3
"""
Visualization Generation Module for Report 001

This module creates visualizations from report card data generated by create_report_card_data.py.
Refactored to match the exact approach from the working generate_report_cards.py.
"""

import sys
import os
import tomllib
import random
from pathlib import Path
from typing import Dict, List, Any, Optional
import pandas as pd
import numpy as np
import plotly.graph_objects as go
from datetime import datetime
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn
import argparse

# Load environment variables from .env file if running locally
try:
    from dotenv import load_dotenv
    # Look for .env file in the project root
    env_path = Path(__file__).parent.parent.parent / '.env'
    if env_path.exists():
        load_dotenv(env_path)
except ImportError:
    # dotenv not available, rely on environment variables being set externally
    pass

# Import our data modules
from create_report_card_data import create_report_card_data
from retrieve_virus_data import retrieve_all_virus_data
from sites_config import get_public_sites, get_private_sites, get_display_name_for_location, get_location_identifier_to_name_mapping


# Environment Detection
def detect_environment() -> str:
    """Detect if running locally or in GitHub Actions."""
    return "github-actions" if os.getenv('GITHUB_ACTIONS') else "local"


def get_output_base_dir(environment: str) -> Path:
    """Get base output directory based on environment."""
    script_dir = Path(__file__).parent
    if environment == "github-actions":
        return script_dir / "github-actions-outputs"
    else:
        return script_dir / "local-outputs"


# Visual constants (matching generate_report_cards.py exactly)
CATEGORY_COLORS = {
    'enteric': '#C70039',
    'respiratory': '#9B59B6', 
    'atypical': '#1F8EFA',
    'animal': '#FFC300'
}

MISSING_FILL = 'rgba(0,0,0,0)'  # Fully transparent

COLORSCALE = [
    [0, 'white'],           # 0 reads per billion (absent)
    [0.1, '#f0f8ff'],       # Very light blue
    [0.3, '#87ceeb'],       # Light blue  
    [0.5, '#4682b4'],       # Medium blue
    [0.7, '#1e90ff'],       # Dodger blue
    [0.9, '#0066cc'],       # Dark blue
    [1.0, '#003366']        # Very dark blue
]


def load_virus_config(config_path: Optional[Path] = None) -> Dict[str, Any]:
    """Load virus configuration from virus.toml file."""
    if config_path is None:
        config_path = Path(__file__).parent / "virus.toml"
    else:
        config_path = Path(config_path)
    
    if not config_path.exists():
        raise FileNotFoundError(f"Virus configuration file not found: {config_path}")
    
    try:
        with open(config_path, 'rb') as f:
            config = tomllib.load(f)
        return config
    except Exception as e:
        raise Exception(f"Failed to load virus configuration from {config_path}: {e}")


# Cache for virus configuration to avoid repeated file loading
_virus_config_cache = None
_virus_category_cache = {}
_trivia_cache = None

# Global fact tracking for each plot to ensure uniqueness
_used_facts_per_plot = {}


def load_trivia_data() -> Dict[str, Any]:
    """Load trivia data from trivia.toml with caching."""
    global _trivia_cache
    
    if _trivia_cache is not None:
        return _trivia_cache
    
    trivia_file = Path(__file__).parent / "trivia.toml"
    if not trivia_file.exists():
        _trivia_cache = {'cities': {}, 'viruses': {}}
        return _trivia_cache
    
    try:
        with open(trivia_file, 'rb') as f:
            _trivia_cache = tomllib.load(f)
        return _trivia_cache
    except Exception as e:
        print(f"Warning: Could not load trivia data: {e}")
        _trivia_cache = {'cities': {}, 'viruses': {}}
        return _trivia_cache


def get_trivia_easter_egg(city_name: str, virus_name: str, cell_index: int, total_cells: int, sequencing_available: bool = True, has_reads: bool = False, plot_id: str = "default") -> Optional[str]:
    """Get trivia Easter egg combining city and virus facts.
    Only places bonus facts in cells where sequencing data is available AND there are reads > 0."""
    
    # Don't place bonus facts in cells without sequencing data OR without reads
    if not sequencing_available or not has_reads:
        return None
        
    # Create a deterministic seed based on cell position and content
    import hashlib
    seed_string = f"{city_name}_{virus_name}_{cell_index}_{total_cells}"
    seed = int(hashlib.md5(seed_string.encode()).hexdigest(), 16) % 1000000
    
    # Set random seed for this specific cell to be deterministic
    random.seed(seed)
    
    # Random probability between 1 in 30 to 1 in 50
    min_prob = 1.0 / 50.0  # 2%
    max_prob = 1.0 / 30.0  # ~3.33%
    probability = min_prob + random.random() * (max_prob - min_prob)
    
    if random.random() > probability:
        return None
    
    trivia_data = load_trivia_data()
    
    # Collect both city and virus facts
    all_facts = []
    
    # Get city facts
    city_key = city_name.lower().replace(', ', '_').replace(' ', '_')
    city_trivia = trivia_data.get('cities', {}).get(city_key, {})
    city_facts = city_trivia.get('facts', [])
    for fact in city_facts:
        all_facts.append(('city', fact))
    
    # Get virus facts
    virus_trivia = trivia_data.get('viruses', {})
    for virus_data in virus_trivia.values():
        if virus_data.get('name') == virus_name:
            virus_facts = virus_data.get('facts', [])
            for fact in virus_facts:
                all_facts.append(('virus', fact))
            break
    
    if not all_facts:
        return None
    
    # Initialize used facts tracking for this plot if not exists
    global _used_facts_per_plot
    if plot_id not in _used_facts_per_plot:
        _used_facts_per_plot[plot_id] = set()
    
    # Find an unused fact
    max_attempts = len(all_facts)
    for attempt in range(max_attempts):
        fact_index = (cell_index + seed + attempt) % len(all_facts)
        fact_type, selected_fact = all_facts[fact_index]
        
        # Create a unique identifier for this fact
        fact_signature = f"{fact_type}:{selected_fact}"
        
        # If this fact hasn't been used in this plot, use it
        if fact_signature not in _used_facts_per_plot[plot_id]:
            _used_facts_per_plot[plot_id].add(fact_signature)
            break
    else:
        # All facts have been used, return None or allow reuse
        return None
    
    # Add unique ID for JavaScript tracking and include entity name
    fact_id = f"fact_{cell_index}_{hashlib.md5(selected_fact.encode()).hexdigest()[:8]}"
    
    # Return the bonus fact formatted for hover text
    # JavaScript will parse this to extract the fact
    entity_name = city_name if fact_type == 'city' else virus_name
    
    # Simple format that survives Plotly's HTML escaping
    # Hide the metadata using CSS for cleaner hover display
    return f"<br><br>🎯 Bonus fact: {selected_fact}<span style='display:none;'>[{fact_type}:{entity_name}:{fact_id}]</span>"


def calculate_bonus_facts_count(city_name: str, virus_name: str, total_cells: int) -> int:
    """Calculate the number of bonus facts for a dashboard with the same logic as get_trivia_easter_egg."""
    # Create a pseudo-random but deterministic seed based on both city and virus names and total cells
    import hashlib
    seed_string = f"{city_name}_{virus_name}_{total_cells}"
    seed = int(hashlib.md5(seed_string.encode()).hexdigest(), 16) % 1000000
    
    # Set random seed for this dashboard to be deterministic but varied
    random.seed(seed)
    
    # Calculate target number of bonus facts for this dashboard
    # Between 1 per 150-250 cells, but at least 1
    min_frequency = 250
    max_frequency = 150
    avg_frequency = (min_frequency + max_frequency) / 2
    target_facts = max(1, int(total_cells / avg_frequency))
    
    return target_facts


def get_virus_category(virus_name: str) -> str:
    """Get virus category from virus.toml configuration with caching."""
    global _virus_config_cache, _virus_category_cache
    
    # Return from cache if available
    if virus_name in _virus_category_cache:
        return _virus_category_cache[virus_name]
    
    # Load config if not cached
    if _virus_config_cache is None:
        _virus_config_cache = load_virus_config()
    
    virus_section = _virus_config_cache.get('virus', {})
    
    for virus_data in virus_section.values():
        if virus_data.get('name') == virus_name:
            category = virus_data.get('type', 'unknown')
            _virus_category_cache[virus_name] = category
            return category
    
    _virus_category_cache[virus_name] = 'unknown'
    return 'unknown'


def get_viruses_by_visibility(visibility: str) -> List[str]:
    """Get viruses by visibility level from virus.toml with caching, sorted by family."""
    global _virus_config_cache
    
    # Load config if not cached
    if _virus_config_cache is None:
        _virus_config_cache = load_virus_config()
    
    virus_section = _virus_config_cache.get('virus', {})
    
    # Collect virus data with family info for sorting
    virus_data_list = []
    for virus_data in virus_section.values():
        virus_name = virus_data.get('name')
        virus_visibility = virus_data.get('visibility')
        virus_family = virus_data.get('family', 'Unknown')
        virus_type = virus_data.get('type', 'unknown')
        
        if not virus_name:
            continue
            
        # Private includes everything, public only includes public viruses
        if visibility == 'private' or virus_visibility == 'public':
            virus_data_list.append({
                'name': virus_name,
                'family': virus_family,
                'type': virus_type
            })
    
    # Sort viruses by family, then by name, with respiratory virus exception
    def virus_sort_key(virus_info):
        name = virus_info['name']
        family = virus_info['family']
        virus_type = virus_info['type']
        
        # Special respiratory viruses go to the bottom
        if virus_type == 'respiratory':
            if ('SARS-CoV-2' in name or 
                'Influenza' in name or 
                'RSV' in name or 'respiratory syncytial' in name.lower()):
                return (99, family, name)  # Force to bottom
        
        # Regular sorting: family first, then name
        return (0, family, name)
    
    sorted_virus_data = sorted(virus_data_list, key=virus_sort_key)
    
    # Extract just the virus names from sorted data
    viruses = [virus_info['name'] for virus_info in sorted_virus_data]
    
    return viruses


def get_sites_by_visibility(visibility: str) -> List[str]:
    """Get sites by visibility level from sites.toml."""
    if visibility == 'public':
        return get_public_sites()
    else:
        return get_private_sites()


def sanitize_name(name: str) -> str:
    """Sanitize names for filenames."""
    import re
    return re.sub(r'[^A-Za-z0-9]+', '_', name)


def prepare_data_for_visualization(df: pd.DataFrame, visibility: str) -> tuple:
    """
    Prepare report card data for visualization by filtering and aggregating.
    Matches the exact approach from generate_report_cards.py.
    """
    console = Console()
    
    if df.empty:
        return df, set()
    
    # Filter viruses by visibility
    allowed_viruses = get_viruses_by_visibility(visibility)
    virus_filtered_df = df[df['virus'].isin(allowed_viruses)]
    
    # Filter sites by visibility  
    allowed_sites = get_sites_by_visibility(visibility)
    site_filtered_df = virus_filtered_df[virus_filtered_df['sewershed_location'].isin(allowed_sites)]
    
    console.print(f"[blue]Data filtering for {visibility}:[/blue]")
    console.print(f"  Original records: {len(df)}")
    console.print(f"  After virus filtering: {len(virus_filtered_df)} (kept {len(allowed_viruses)} viruses)")
    console.print(f"  After site filtering: {len(site_filtered_df)} (kept {len(allowed_sites)} sites)")
    
    # Calculate reads per billion if not already present
    if 'reads_per_billion' not in site_filtered_df.columns:
        site_filtered_df = site_filtered_df.copy()
        site_filtered_df['reads_per_billion'] = 0.0
        mask = (site_filtered_df['total_reads'] > 0) & (site_filtered_df['mapped_reads'] >= 0)
        site_filtered_df.loc[mask, 'reads_per_billion'] = (
            site_filtered_df.loc[mask, 'mapped_reads'].astype(float) / 
            site_filtered_df.loc[mask, 'total_reads'].astype(float)
        ) * 1e9
    
    # Convert time_label to datetime and add week aggregation (matching generate_report_cards.py)
    site_filtered_df['sample_date'] = pd.to_datetime(site_filtered_df['time_label'], errors='coerce')
    
    # Add week aggregation exactly like generate_report_cards.py
    if site_filtered_df['sample_date'].notna().any():
        site_filtered_df['week_start'] = site_filtered_df['sample_date'].dt.to_period('W-SAT').dt.start_time
        site_filtered_df['week_label'] = site_filtered_df['week_start'].dt.strftime('%Y-%m-%d')
        site_filtered_df['week_date'] = pd.to_datetime(site_filtered_df['week_label'], errors='coerce')
    else:
        console.print("[yellow]No valid sample dates found[/yellow]")
        return pd.DataFrame(), set()
    
    # Create sequencing coverage set (matching generate_report_cards.py approach)
    sequencing_coverage = set()
    valid_coverage = site_filtered_df.dropna(subset=['sewershed_location', 'week_label'])
    sequencing_coverage = set(zip(valid_coverage['sewershed_location'], valid_coverage['week_label']))
    
    # Aggregate by week, location, and virus (matching generate_report_cards.py)
    agg_cols = ['sewershed_location', 'virus', 'week_label']
    agg_dict = {
        'mapped_reads': 'sum',
        'total_reads': 'first',  # Take first total reads (represents sequencing depth)
        'reads_per_billion': 'sum',  # Sum the normalized values
        'sample_date': 'first',
        'week_start': 'first',
        'week_date': 'first'
    }
    
    agg_df = site_filtered_df.groupby(agg_cols).agg(agg_dict).reset_index()
    
    console.print(f"[green]✓ Aggregated to {len(agg_df)} records[/green]")
    
    return agg_df, sequencing_coverage


def create_by_city_heatmap(df: pd.DataFrame, city_name: str, virus_category: str, 
                          output_dir: Path, visibility: str, sequencing_coverage: set, test_mode: bool = False, generate_html: bool = True) -> bool:
    """
    Create a heatmap for a specific city and virus category.
    Optimized version with pre-filtering and vectorized operations.
    """
    console = Console()
    
    # Pre-filter data (Optimization 4)
    city_df = df[df['sewershed_location'] == city_name]
    
    if city_df.empty:
        display_name = get_display_name_for_location(city_name)
        console.print(f"[yellow]No data found for city: {display_name}[/yellow]")
        return False
    
    # Get viruses for this category
    all_viruses = get_viruses_by_visibility(visibility)
    category_viruses = [v for v in all_viruses if get_virus_category(v) == virus_category]
    
    if not category_viruses:
        console.print(f"[yellow]No {virus_category} viruses found for {city_name}[/yellow]")
        return False
    
    # Pre-filter to category viruses and create indexed DataFrame for fast lookups
    category_df = city_df[city_df['virus'].isin(category_viruses)].copy()
    
    # Get weeks where sequencing was performed for this city using set operations
    city_coverage_weeks = {week_label for location, week_label in sequencing_coverage if location == city_name}
    
    if city_coverage_weeks:
        # Convert to datetime once
        city_weeks_dt = pd.to_datetime(list(city_coverage_weeks))
        earliest_seq_date = city_weeks_dt.min()
        latest_seq_date = city_weeks_dt.max()
        all_weeks = pd.date_range(start=earliest_seq_date, end=latest_seq_date, freq='W').tolist()
    else:
        # Fallback to data available for this city
        all_weeks = sorted(category_df['week_date'].dropna().unique())
    
    if not all_weeks:
        console.print(f"[yellow]No valid weeks found for {city_name} {virus_category}[/yellow]")
        return False
    
    console.print(f"[cyan]Processing {city_name} {virus_category}: {len(category_viruses)} viruses, {len(all_weeks)} weeks[/cyan]")
    
    # Pre-create week labels for faster lookup
    week_labels = [week.strftime('%Y-%m-%d') for week in all_weeks]
    
    # Vectorized data matrix creation (Optimization 5)
    n_viruses = len(category_viruses)
    n_weeks = len(all_weeks)
    
    # Initialize arrays
    z_values = np.full((n_viruses, n_weeks), np.nan)
    hover_text = [['' for _ in range(n_weeks)] for _ in range(n_viruses)]
    
    # Create pivot table for faster lookups
    if not category_df.empty:
        # Set multi-index for efficient lookups
        category_df.set_index(['virus', 'week_date'], inplace=True)
        category_df.sort_index(inplace=True)
    
    # Process each virus
    for i, virus in enumerate(category_viruses):
        for j, (week, week_label) in enumerate(zip(all_weeks, week_labels)):
            sequencing_available = (city_name, week_label) in sequencing_coverage
            
            # Try to get data from indexed DataFrame
            try:
                if not category_df.empty and (virus, week) in category_df.index:
                    row_data = category_df.loc[(virus, week)]
                    if isinstance(row_data, pd.DataFrame):
                        row_data = row_data.iloc[0]
                    
                    reads_per_billion = row_data['reads_per_billion']
                    mapped_reads = row_data['mapped_reads']
                    total_reads = row_data['total_reads']
                    
                    if reads_per_billion > 0:
                        z_values[i, j] = np.log10(reads_per_billion + 1)
                    else:
                        z_values[i, j] = 0
                    
                    # Generate cell index for Easter egg calculation
                    cell_index = i * len(all_weeks) + j
                    base_hover = (
                        f"<b>{virus}</b><br>"
                        f"Week: {week_label}<br>"
                        f"Mapped reads: {mapped_reads:,.0f}<br>"
                        f"Total reads: {total_reads:,.0f}<br>"
                        f"Reads per billion: {reads_per_billion:.1f}"
                    )
                    
                    # Check for trivia Easter egg
                    # Create plot ID for this city-category combination
                    plot_id = f"city_{city_name}_{virus_category}"
                    easter_egg = get_trivia_easter_egg(city_name, virus, cell_index, len(category_viruses) * len(all_weeks), sequencing_available=True, has_reads=True, plot_id=plot_id)
                    if easter_egg:
                        hover_text[i][j] = base_hover + easter_egg
                    else:
                        hover_text[i][j] = base_hover
                elif sequencing_available:
                    z_values[i, j] = 0
                    cell_index = i * len(all_weeks) + j
                    base_hover = f"<b>{virus}</b><br>Week: {week_label}<br>Sequencing performed, 0 reads detected"
                    # No bonus facts for zero read cells
                    hover_text[i][j] = base_hover
                else:
                    cell_index = i * len(all_weeks) + j
                    base_hover = f"<b>{virus}</b><br>Week: {week_label}<br>No sequencing data available"
                    # No bonus facts in cells without sequencing data
                    hover_text[i][j] = base_hover
            except KeyError:
                if sequencing_available:
                    z_values[i, j] = 0
                    cell_index = i * len(all_weeks) + j
                    base_hover = f"<b>{virus}</b><br>Week: {week_label}<br>Sequencing performed, 0 reads detected"
                    # No bonus facts for zero read cells
                    hover_text[i][j] = base_hover
                else:
                    cell_index = i * len(all_weeks) + j
                    base_hover = f"<b>{virus}</b><br>Week: {week_label}<br>No sequencing data available"
                    # No bonus facts in cells without sequencing data
                    hover_text[i][j] = base_hover
    
    # Memory optimization (Optimization 6) - clear the indexed DataFrame
    del category_df
    
    # Calculate zmax from actual data for proper color scaling
    computed_z_max = np.nanmax(z_values) if not np.isnan(z_values).all() else 1.0
    z_max = max(1.0, computed_z_max)  # Ensure minimum zmax is 1
    
    # Convert timestamps to datetime for plotly
    x_dates = [week.to_pydatetime() for week in all_weeks]
    
    # Create heatmap
    fig = go.Figure(data=go.Heatmap(
        z=z_values.tolist(),  # Convert numpy array to list
        x=x_dates,
        y=category_viruses,
        text=hover_text,
        hovertemplate='%{text}<extra></extra>',
        colorscale=COLORSCALE,
        zmin=0,
        zmax=z_max,
        zauto=False,  # Disable auto-scaling to enforce zmin/zmax
        showscale=True,
        colorbar=dict(
            title=dict(text="Reads per Billion<br>(log₁₀ scale)", font=dict(size=14)),
            tickfont=dict(size=12)
        ),
        connectgaps=False,
        xgap=1,
        ygap=1
    ))
    
    # Smart missing data detection - only within active range
    # First, determine the active date range for this city
    city_coverage_weeks_str = {week for loc, week in sequencing_coverage if loc == city_name}
    
    if city_coverage_weeks_str:
        # Convert to datetime to find range
        city_coverage_weeks_dt = pd.to_datetime(list(city_coverage_weeks_str))
        earliest_coverage = city_coverage_weeks_dt.min()
        latest_coverage = city_coverage_weeks_dt.max()
        
        # Find indices of weeks within the coverage range
        active_week_indices = []
        for j, week in enumerate(all_weeks):
            if earliest_coverage <= week <= latest_coverage:
                active_week_indices.append(j)
        
        # Only check for missing data within the active range
        shapes = []
        for i in range(len(category_viruses)):
            for j in active_week_indices:
                if np.isnan(z_values[i, j]):
                    shapes.append(dict(
                        type="rect",
                        x0=j-0.5, x1=j+0.5,
                        y0=i-0.5, y1=i+0.5,
                        fillcolor=MISSING_FILL,
                        line=dict(color=MISSING_FILL, width=0),
                        layer="above"
                    ))
    else:
        # No coverage for this city, no shapes needed
        shapes = []
    
    # Calculate actual data range for this city
    # Use the pre-computed week range
    if len(all_weeks) > 0:
        earliest_data_date = all_weeks[0]
        latest_data_date = all_weeks[-1]
    else:
        earliest_data_date = None
        latest_data_date = None
    
    # Calculate bonus facts for X-axis label using the new algorithm
    total_cells = len(category_viruses) * len(all_weeks)
    # Use the first virus in the category as representative for calculation
    representative_virus = category_viruses[0] if category_viruses else "Unknown"
    bonus_facts_count = calculate_bonus_facts_count(city_name, representative_virus, total_cells)
    # Initially use empty title, game title will be set when activated
    x_axis_title = ""
    
    # Configure layout (matching generate_report_cards.py exactly)
    layout_settings = dict(
        xaxis=dict(
            title=dict(text=x_axis_title, font=dict(size=14)),
            tickfont=dict(size=9),
            tickangle=45,
            tickmode='auto',
            nticks=20,  # Limit number of ticks to prevent overcrowding
            range=[earliest_data_date, latest_data_date] if earliest_data_date and latest_data_date else None
        ),
        yaxis=dict(
            title=dict(text='Virus', font=dict(size=14)),
            tickfont=dict(size=10)
        ),
        width=1200,  # Fixed width to fit in 1200 pixels
        height=max(400, len(category_viruses) * 30),
        margin=dict(l=200, r=100, t=100, b=100)
    )
    
    # Configure title and layout
    # Get display name for the city
    display_name = get_display_name_for_location(city_name)
    title_text = f'<b style="color:black">{display_name}</b>'
    
    fig.update_layout(
        title=dict(
            text=title_text,
            font=dict(size=18, color='black'),
            x=0.5
        ),
        shapes=shapes,  # Add all shapes at once
        **layout_settings
    )
    
    # Save files
    safe_city_name = sanitize_name(city_name)
    html_path = output_dir / 'by_city' / f'{safe_city_name}_{virus_category}.html'
    png_path = output_dir / 'by_city' / f'{safe_city_name}_{virus_category}.png'
    
    # Always save PNG
    fig.write_image(png_path, width=1200, height=fig.layout.height, scale=2)
    # Only save HTML if requested and not in test mode
    if not test_mode and generate_html:
        write_html_with_bonus_tracking(fig, html_path, bonus_facts_count)
    
    console.print(f"[green]✓ Created {virus_category} heatmap for {display_name}[/green]")
    return True


def create_by_virus_heatmap(df: pd.DataFrame, virus_name: str, 
                           output_dir: Path, visibility: str, sequencing_coverage: set, test_mode: bool = False, generate_html: bool = True) -> bool:
    """
    Create a heatmap for a specific virus across all cities.
    Highly optimized version with vectorized operations.
    """
    console = Console()
    
    # Pre-filter data for this virus (Optimization 4)
    virus_df = df[df['virus'] == virus_name]
    
    # Early exit if no data
    if virus_df.empty:
        console.print(f"[yellow]No data found for virus: {virus_name}[/yellow]")
        return False
    
    # Get ALL weeks where sequencing was performed using set operations
    all_coverage_weeks = {week_label for _, week_label in sequencing_coverage}
    
    if all_coverage_weeks:
        # Convert to datetime once
        weeks_dt = pd.to_datetime(list(all_coverage_weeks))
        earliest_seq_date = weeks_dt.min()
        latest_seq_date = weeks_dt.max()
        all_weeks = pd.date_range(start=earliest_seq_date, end=latest_seq_date, freq='W').tolist()
    else:
        all_weeks = []
    
    # Get all cities for this visibility level
    all_cities = get_sites_by_visibility(visibility)
    
    if not all_weeks:
        console.print(f"[yellow]No valid weeks found for visualization[/yellow]")
        return False
    
    console.print(f"[cyan]Processing {virus_name}: {len(all_cities)} cities, {len(all_weeks)} weeks[/cyan]")
    
    # Pre-create week labels for faster lookup
    week_labels = [week.strftime('%Y-%m-%d') for week in all_weeks]
    
    # Vectorized data matrix creation (Optimization 5)
    n_cities = len(all_cities)
    n_weeks = len(all_weeks)
    
    # Initialize arrays
    z_values = np.full((n_cities, n_weeks), np.nan)
    hover_text = [['' for _ in range(n_weeks)] for _ in range(n_cities)]
    
    # Create pivot table for much faster access (major optimization)
    if not virus_df.empty:
        # Pivot the data to have cities as rows and weeks as columns
        pivot_df = virus_df.pivot_table(
            index='sewershed_location',
            columns='week_date',
            values=['reads_per_billion', 'mapped_reads', 'total_reads'],
            aggfunc='first'
        )
        
        # Process data using vectorized operations
        for i, city in enumerate(all_cities):
            for j, (week, week_label) in enumerate(zip(all_weeks, week_labels)):
                sequencing_available = (city, week_label) in sequencing_coverage
                
                # Check if we have data for this city/week
                has_data = False
                if city in pivot_df.index and week in pivot_df.columns.get_level_values(1):
                    try:
                        reads_per_billion = pivot_df.loc[city, ('reads_per_billion', week)]
                        if pd.notna(reads_per_billion):
                            has_data = True
                            mapped_reads = pivot_df.loc[city, ('mapped_reads', week)]
                            total_reads = pivot_df.loc[city, ('total_reads', week)]
                            
                            if reads_per_billion > 0:
                                z_values[i, j] = np.log10(reads_per_billion + 1)
                            else:
                                z_values[i, j] = 0
                            
                            # Generate cell index for Easter egg calculation
                            cell_index = i * len(week_labels) + j
                            # Get display name for hover text
                            city_display_name = get_display_name_for_location(city)
                            base_hover = (
                                f"<b>{city_display_name}</b><br>"
                                f"Week: {week_label}<br>"
                                f"Mapped reads: {mapped_reads:,.0f}<br>"
                                f"Total reads: {total_reads:,.0f}<br>"
                                f"Reads per billion: {reads_per_billion:.1f}"
                            )
                            
                            # Check for trivia Easter egg
                            easter_egg = get_trivia_easter_egg(city, virus_name, cell_index, len(all_cities) * len(week_labels), sequencing_available=True, has_reads=True, plot_id=f"virus_{virus_name}")
                            if easter_egg:
                                hover_text[i][j] = base_hover + easter_egg
                            else:
                                hover_text[i][j] = base_hover
                    except (KeyError, IndexError):
                        pass
                
                if not has_data:
                    cell_index = i * len(week_labels) + j
                    if sequencing_available:
                        z_values[i, j] = 0
                        city_display_name = get_display_name_for_location(city)
                        base_hover = f"<b>{city_display_name}</b><br>Week: {week_label}<br>Sequencing performed, 0 reads detected"
                        # No bonus facts for zero read cells
                        hover_text[i][j] = base_hover
                    else:
                        city_display_name = get_display_name_for_location(city)
                        base_hover = f"<b>{city_display_name}</b><br>Week: {week_label}<br>No sequencing data available"
                        # No bonus facts in cells without sequencing data
                        hover_text[i][j] = base_hover
    else:
        # No data for this virus - fill with appropriate values
        for i, city in enumerate(all_cities):
            for j, week_label in enumerate(week_labels):
                cell_index = i * len(week_labels) + j
                if (city, week_label) in sequencing_coverage:
                    z_values[i, j] = 0
                    base_hover = f"<b>{city}</b><br>Week: {week_label}<br>Sequencing performed, 0 reads detected"
                    # No bonus facts in cells without sequencing data
                    hover_text[i][j] = base_hover
                else:
                    base_hover = f"<b>{city}</b><br>Week: {week_label}<br>No sequencing data available"
                    # No bonus facts in cells without sequencing data
                    hover_text[i][j] = base_hover
    
    # Memory optimization (Optimization 6) - clear the indexed DataFrame
    del virus_df
    
    # Calculate zmax from actual data for proper color scaling
    computed_z_max = np.nanmax(z_values) if not np.isnan(z_values).all() else 1.0
    z_max = max(1.0, computed_z_max)  # Ensure minimum zmax is 1
    
    # Convert timestamps to datetime for plotly
    x_dates = [week.to_pydatetime() for week in all_weeks]
    
    # Get display names for cities on y-axis
    city_display_names = [get_display_name_for_location(city) for city in all_cities]
    
    # Create heatmap
    fig = go.Figure(data=go.Heatmap(
        z=z_values.tolist(),  # Convert numpy array to list
        x=x_dates,
        y=city_display_names,
        text=hover_text,
        hovertemplate='%{text}<extra></extra>',
        colorscale=COLORSCALE,
        zmin=0,
        zmax=z_max,
        zauto=False,  # Disable auto-scaling to enforce zmin/zmax
        showscale=True,
        colorbar=dict(
            title=dict(text="Reads per Billion<br>(log₁₀ scale)", font=dict(size=14)),
            tickfont=dict(size=12)
        ),
        connectgaps=False,
        xgap=1,
        ygap=1
    ))
    
    # Smart missing data detection - pre-compute city date ranges
    city_ranges = {}
    for city in all_cities:
        city_weeks_str = {week for loc, week in sequencing_coverage if loc == city}
        if city_weeks_str:
            city_weeks_dt = pd.to_datetime(list(city_weeks_str))
            city_ranges[city] = (city_weeks_dt.min(), city_weeks_dt.max())
    
    # Create shapes for missing data within active ranges only
    shapes = []
    for i, city in enumerate(all_cities):
        if city in city_ranges:
            min_date, max_date = city_ranges[city]
            for j, week in enumerate(all_weeks):
                # Only mark as missing if within the city's active range
                if min_date <= week <= max_date and np.isnan(z_values[i, j]):
                    shapes.append(dict(
                        type="rect",
                        x0=j-0.5, x1=j+0.5,
                        y0=i-0.5, y1=i+0.5,
                        fillcolor=MISSING_FILL,
                        line=dict(color=MISSING_FILL, width=0),
                        layer="above"
                    ))
    
    # Calculate actual data range for this virus
    # Use the pre-computed week range
    if len(all_weeks) > 0:
        earliest_data_date = all_weeks[0]
        latest_data_date = all_weeks[-1]
    else:
        earliest_data_date = None
        latest_data_date = None
    
    # Calculate bonus facts for X-axis label using the new algorithm
    total_cells = len(all_cities) * len(all_weeks)
    # Use the first city as representative for calculation
    representative_city = all_cities[0] if all_cities else "Unknown"
    bonus_facts_count = calculate_bonus_facts_count(representative_city, virus_name, total_cells)
    # Initially use empty title, game title will be set when activated
    x_axis_title = ""
    
    # Configure layout (matching generate_report_cards.py exactly)
    layout_settings = dict(
        xaxis=dict(
            title=dict(text=x_axis_title, font=dict(size=14)),
            tickfont=dict(size=9),
            tickangle=45,
            tickmode='auto',
            nticks=20,  # Limit number of ticks to prevent overcrowding
            range=[earliest_data_date, latest_data_date] if earliest_data_date and latest_data_date else None
        ),
        yaxis=dict(
            title=dict(text='City', font=dict(size=14)),
            tickfont=dict(size=10)
        ),
        width=1200,  # Fixed width to fit in 1200 pixels
        height=max(400, len(all_cities) * 50),
        margin=dict(l=200, r=100, t=100, b=100)
    )
    
    # Configure title and layout
    fig.update_layout(
        title=dict(
            text=f'<b style="color:black">{virus_name}</b>',
            font=dict(size=18, color='black'),
            x=0.5
        ),
        shapes=shapes,  # Add all shapes at once
        **layout_settings
    )
    
    # Save files
    safe_virus_name = sanitize_name(virus_name)
    html_path = output_dir / 'by_virus' / f'{safe_virus_name}.html'
    png_path = output_dir / 'by_virus' / f'{safe_virus_name}.png'
    
    # Always save PNG
    fig.write_image(png_path, width=1200, height=fig.layout.height, scale=2)
    # Only save HTML if requested and not in test mode
    if not test_mode and generate_html:
        write_html_with_bonus_tracking(fig, html_path, bonus_facts_count)
    
    virus_category = get_virus_category(virus_name)
    console.print(f"[green]✓ Created heatmap for {virus_name} ({virus_category})[/green]")
    return True


def write_html_with_bonus_tracking(fig, html_path: Path, total_bonus_facts: int):
    """Write HTML file with JavaScript bonus fact tracking and celebration."""
    
    # Generate the standard Plotly HTML
    html_string = fig.to_html(include_plotlyjs=True)
    
    # Create unique iframe identifier to completely isolate each instance
    iframe_id = f"iframe_{abs(hash(str(html_path)))}"
    
    # JavaScript code for bonus fact tracking
    bonus_tracking_js = f"""
<script>
// Virus Dashboard Games - Completely isolated for iframe usage
(function() {{
    // Create completely unique namespace for this specific iframe instance
    const iframeId = '{iframe_id}';
    const gameState = {{
        iframeId: iframeId,
        discoveredFacts: new Set(),
        totalBonusFacts: {total_bonus_facts},
        factsList: [],
        startTime: Date.now(),
        timerInterval: null,
        gameActivated: false,
        currentGame: null
    }};
    
    // Store in completely isolated namespace
    if (!window.virusDashboardGames) {{
        window.virusDashboardGames = {{}};
    }}
    window.virusDashboardGames[iframeId] = gameState;

    // Function definitions (scoped to this iframe)

    function showGameMenu() {{
        // Create game selection menu
        const menuOverlay = document.createElement('div');
        menuOverlay.id = `game-menu-${{iframeId}}`;
        
        // Check if we're in game mode (opened in new window)
        const urlParams = new URLSearchParams(window.location.search);
        const isGameMode = urlParams.get('game') === 'true';
        
        if (isGameMode) {{
            // In game mode, show menu as primary content without overlay
            menuOverlay.style.cssText = `
                position: fixed;
                top: 50%;
                left: 50%;
                transform: translate(-50%, -50%);
                z-index: 9999;
                font-family: Arial, sans-serif;
            `;
        }} else {{
            // Normal overlay mode
            menuOverlay.style.cssText = `
                position: absolute;
                top: 0;
                left: 0;
                width: 100%;
                height: 100%;
                background: rgba(0, 0, 0, 0.7);
                z-index: 9999;
                display: flex;
                justify-content: center;
                align-items: center;
                font-family: Arial, sans-serif;
            `;
        }}
        
        const menuContainer = document.createElement('div');
        menuContainer.style.cssText = `
            background: white;
            border-radius: 15px;
            padding: 30px;
            box-shadow: 0 8px 16px rgba(0,0,0,0.2);
            max-width: 500px;
            text-align: center;
        `;
        
        menuContainer.innerHTML = `
            <h2 style="margin: 0 0 20px 0; color: #333; font-size: 24px;">🎮 Virus Dashboard Games</h2>
            <p style="margin: 0 0 25px 0; color: #666; font-size: 16px;">Choose your adventure in virus surveillance data!</p>
            
            <div style="margin: 15px 0;">
                <button id="triviradae-${{iframeId}}" style="
                    background: linear-gradient(135deg, #667eea 0%, #764ba2 100%);
                    color: white;
                    border: none;
                    padding: 15px 30px;
                    border-radius: 25px;
                    font-size: 16px;
                    font-weight: bold;
                    cursor: pointer;
                    margin: 10px;
                    min-width: 200px;
                    box-shadow: 0 4px 8px rgba(0,0,0,0.2);
                    transition: transform 0.2s ease;
                " onmouseover="this.style.transform='scale(1.05)'" onmouseout="this.style.transform='scale(1)'">
                    🎯 Triviradae
                </button>
                <div style="font-size: 12px; color: #888; margin-top: 5px;">Hunt for hidden bonus facts in the data!</div>
            </div>
            
            <div style="margin: 15px 0;">
                <button id="outbreak-${{iframeId}}" style="
                    background: linear-gradient(135deg, #f093fb 0%, #f5576c 100%);
                    color: white;
                    border: none;
                    padding: 15px 30px;
                    border-radius: 25px;
                    font-size: 16px;
                    font-weight: bold;
                    cursor: pointer;
                    margin: 10px;
                    min-width: 200px;
                    box-shadow: 0 4px 8px rgba(0,0,0,0.2);
                    transition: transform 0.2s ease;
                " onmouseover="this.style.transform='scale(1.05)'" onmouseout="this.style.transform='scale(1)'">
                    🦠 Outbreak
                </button>
                <div style="font-size: 12px; color: #888; margin-top: 5px;">Destroy wastewater viruses before time runs out!</div>
            </div>
            
            <button id="close-menu-${{iframeId}}" style="
                background: #ccc;
                color: #666;
                border: none;
                padding: 10px 20px;
                border-radius: 15px;
                font-size: 14px;
                cursor: pointer;
                margin-top: 20px;
            ">Cancel</button>
        `;
        
        menuOverlay.appendChild(menuContainer);
        document.body.appendChild(menuOverlay);
        
        // Add event listeners
        document.getElementById(`triviradae-${{iframeId}}`).addEventListener('click', function() {{
            document.body.removeChild(menuOverlay);
            activateTriviradae();
        }});
        
        document.getElementById(`outbreak-${{iframeId}}`).addEventListener('click', function() {{
            // Open outbreak game in new window
            window.open('../../outbreak_game.html', '_blank', 'width=1200,height=800,resizable=yes,scrollbars=yes');
        }});
        
        document.getElementById(`close-menu-${{iframeId}}`).addEventListener('click', function() {{
            if (isGameMode) {{
                // In game mode, close the window
                window.close();
            }} else {{
                // In normal mode, just remove the overlay
                document.body.removeChild(menuOverlay);
            }}
        }});
        
        // Close menu if clicking outside (only in non-game mode)
        if (!isGameMode) {{
            menuOverlay.addEventListener('click', function(e) {{
                if (e.target === menuOverlay) {{
                    document.body.removeChild(menuOverlay);
                }}
            }});
        }}
    }}

    function activateTriviradae() {{
        const state = window.virusDashboardGames[iframeId];
        state.currentGame = 'triviradae';
        state.gameActivated = true;
        state.startTime = Date.now(); // Reset start time when game actually starts
        
        // Show the facts display
        const factsContainer = document.getElementById(`discovered-facts-${{iframeId}}`);
        if (factsContainer) {{
            factsContainer.style.display = 'block';
        }}
        
        
        // Update X-axis title to show game instructions
        updateXAxisTitle();
        
        // Start the game
        setupBonusFactTracking();
        startTimer();
    }}

    function hideGameElements() {{
        // Hide the facts display initially
        const factsContainer = document.getElementById(`discovered-facts-${{iframeId}}`);
        if (factsContainer) {{
            factsContainer.style.display = 'none';
        }}
        
        // X-axis labels (dates) should remain visible at all times
        // Only the X-axis title needs to be updated when game is activated
    }}
    
    // Function to properly return to main display
    function returnToDisplay() {{
        // Use document instead of parent.document or window.parent.document
        const overlay = document.getElementById(`celebration-${{iframeId}}`);
        if (overlay && overlay.parentNode) {{
            overlay.parentNode.removeChild(overlay);
        }}
        
        // Hide facts display
        const factsContainer = document.getElementById(`discovered-facts-${{iframeId}}`);
        if (factsContainer) {{
            factsContainer.style.display = 'none';
        }}
        
        // Reset game state
        const state = window.virusDashboardGames[iframeId];
        state.gameActivated = false;
        state.currentGame = null;
        state.discoveredFacts.clear();
        state.factsList = [];
        
        // Stop timer
        if (state.timerInterval) {{
            clearInterval(state.timerInterval);
            state.timerInterval = null;
        }}
        
        
        // Reset X-axis title
        const plotDiv = document.querySelector('.plotly-graph-div');
        if (plotDiv && plotDiv.data) {{
            Plotly.relayout(plotDiv, {{
                'xaxis.title.text': ''
            }});
        }}
    }}

    // Removed scanExistingFacts function - it was causing all facts to be found immediately

    function updateXAxisTitle() {{
        // Find the plotly div and update the x-axis title
        const plotDiv = document.querySelector('.plotly-graph-div');
        if (plotDiv && plotDiv.data) {{
            // Get the current bonus facts count from the page title or calculate it
            const bonusFacts = {total_bonus_facts};
            const newTitle = `Can you find the ${{bonusFacts}} bonus facts hidden in this dashboard?`;
            
            Plotly.relayout(plotDiv, {{
                'xaxis.title.text': newTitle
            }});
        }}
    }}

    function createFactsDisplay() {{
        const body = document.body;
        
        // Create container for discovered facts
        const factsContainer = document.createElement('div');
        factsContainer.id = `discovered-facts-${{iframeId}}`;
        factsContainer.style.cssText = `
            position: absolute;
            bottom: 0px;
            left: 20px;
            max-width: 400px;
            background: white;
            border: 2px solid #0066cc;
            border-radius: 10px;
            padding: 15px;
            box-shadow: 0 4px 8px rgba(0,0,0,0.1);
            z-index: 1000;
            font-family: Arial, sans-serif;
            font-size: 14px;
            max-height: 400px;
            overflow-y: auto;
        `;
        
        // Timer display
        const timerDiv = document.createElement('div');
        timerDiv.id = `timer-display-${{iframeId}}`;
        timerDiv.style.cssText = 'margin-bottom: 10px; color: #666; font-size: 12px;';
        timerDiv.textContent = 'Time: 0:00';
        factsContainer.appendChild(timerDiv);
    
        const title = document.createElement('h3');
        title.textContent = `Discovered Bonus Facts (0/{total_bonus_facts})`;
        title.style.cssText = 'margin: 0 0 10px 0; color: #0066cc; font-size: 16px;';
        factsContainer.appendChild(title);
        
        const factsList = document.createElement('ul');
        factsList.id = `facts-list-${{iframeId}}`;
        factsList.style.cssText = 'margin: 0; padding-left: 20px; list-style-type: none;';
        factsContainer.appendChild(factsList);
        
        body.appendChild(factsContainer);
    }}

    function startTimer() {{
        const state = window.virusDashboardGames[iframeId];
        state.timerInterval = setInterval(updateTimer, 1000);
    }}

    function updateTimer() {{
        const state = window.virusDashboardGames[iframeId];
        const elapsed = Math.floor((Date.now() - state.startTime) / 1000);
        const minutes = Math.floor(elapsed / 60);
        const seconds = elapsed % 60;
        const timerDisplay = document.getElementById(`timer-display-${{iframeId}}`);
        if (timerDisplay) {{
            timerDisplay.textContent = `Time: ${{minutes}}:${{seconds.toString().padStart(2, '0')}}`;
        }}
    }}

    function setupBonusFactTracking() {{
        const state = window.virusDashboardGames[iframeId];
        
        // Enhanced detection using multiple strategies
        
        // Strategy 1: Mutation Observer for DOM changes
        const observer = new MutationObserver(function(mutations) {{
            let foundNewFacts = false;
            mutations.forEach(function(mutation) {{
                mutation.addedNodes.forEach(function(node) {{
                    if (node.nodeType === 1) {{ // Element node
                        const bonusFacts = node.querySelectorAll ? node.querySelectorAll('.bonus-fact') : [];
                        bonusFacts.forEach(function(fact) {{
                            if (trackBonusFact(fact)) {{
                                foundNewFacts = true;
                            }}
                        }});
                    }}
                }});
            }});
        }});
        
        observer.observe(document.body, {{
            childList: true,
            subtree: true
        }});
        
        // Strategy 2: Periodic scanning for bonus facts in visible tooltips only
        setInterval(function() {{
            // Only look for facts in Plotly tooltips that are currently visible
            const tooltips = document.querySelectorAll('.hoverlayer .hovertext');
            tooltips.forEach(function(tooltip) {{
                if (tooltip.style.display !== 'none' && tooltip.style.opacity !== '0') {{
                    const tooltipText = tooltip.textContent || tooltip.innerText || '';
                    const bonusFactPattern = /🎯\\s*Bonus fact:\\s*([^\\[]+)\\[([^:]+):([^:]+):([^\\]]+)\\]/;
                    const match = tooltipText.match(bonusFactPattern);
                    
                    if (match) {{
                        const factText = match[1].trim();
                        const factType = match[2];
                        const entityName = match[3];
                        const factId = match[4];
                        
                        if (!state.discoveredFacts.has(factId)) {{
                            state.discoveredFacts.add(factId);
                            // Add fact with entity name prefix
                            const formattedFact = `<strong>${{entityName}}:</strong> ${{factText}}`;
                            state.factsList.push(formattedFact);
                            updateFactsDisplay();
                            
                            // Check if all facts found
                            if (state.discoveredFacts.size >= state.totalBonusFacts) {{
                                celebrate();
                            }}
                        }}
                    }}
                }}
            }});
        }}, 100); // Check more frequently but only visible tooltips
    
        // Strategy 3: Event listeners for hover and click events
        document.addEventListener('mouseover', function(e) {{
            setTimeout(() => {{
                scanForBonusFacts();
            }}, 200);
        }});
        
        document.addEventListener('click', function(e) {{
            setTimeout(() => {{
                scanForBonusFacts();
            }}, 200);
        }});
    }}

    function scanForBonusFacts() {{
        const state = window.virusDashboardGames[iframeId];
        // Only scan visible Plotly tooltips
        const tooltips = document.querySelectorAll('.hoverlayer .hovertext');
        tooltips.forEach(function(tooltip) {{
            if (tooltip.style.display !== 'none' && tooltip.style.opacity !== '0') {{
                const text = tooltip.textContent || tooltip.innerText || '';
                if (text.includes('🎯') && text.includes('Bonus fact:')) {{
                    const pattern = /🎯\\s*Bonus fact:\\s*([^\\[]+)\\[([^:]+):([^:]+):([^\\]]+)\\]/;
                    const match = text.match(pattern);
                    
                    if (match) {{
                        const factText = match[1].trim();
                        const factType = match[2];
                        const entityName = match[3];
                        const factId = match[4];
                        
                        if (!state.discoveredFacts.has(factId)) {{
                            state.discoveredFacts.add(factId);
                            const formattedFact = `<strong>${{entityName}}:</strong> ${{factText}}`;
                            state.factsList.push(formattedFact);
                            updateFactsDisplay();
                            
                            if (state.discoveredFacts.size >= state.totalBonusFacts) {{
                                celebrate();
                            }}
                        }}
                    }}
                }}
            }}
        }});
    }}

    function trackBonusFact(factElement) {{
        const state = window.virusDashboardGames[iframeId];
        if (!factElement || !factElement.id) return false;
        
        if (state.discoveredFacts.has(factElement.id)) return false;
        
        const factId = factElement.id;
        const factText = factElement.textContent || factElement.innerText;
        
        if (!factText) return false;
        
        // Mark as discovered
        state.discoveredFacts.add(factId);
        state.factsList.push(factText);
        
        // Update display
        updateFactsDisplay();
        
        // Add visual feedback
        if (factElement.style) {{
            factElement.style.backgroundColor = '#ffff99';
            factElement.style.fontWeight = 'bold';
        }}
        
        // Use checkWinCondition instead of direct celebrate
        checkWinCondition();
        
        return true;
    }}

    // Add checkWinCondition function with minimum game time
    function checkWinCondition() {{
        const state = window.virusDashboardGames[iframeId];
        
        // Only check win condition if game has been active for at least 2 seconds
        const gameActiveTime = Date.now() - state.startTime;
        if (gameActiveTime < 2000) {{
            // Schedule another check in 1 second
            setTimeout(checkWinCondition, 1000);
            return;
        }}
        
        // Check if all facts have been found
        if (state.discoveredFacts.size >= state.totalBonusFacts) {{
            celebrate();
        }}
    }}

    function updateFactsDisplay() {{
        const state = window.virusDashboardGames[iframeId];
        const title = document.querySelector(`#discovered-facts-${{iframeId}} h3`);
        const list = document.getElementById(`facts-list-${{iframeId}}`);
        
        if (title) {{
            title.textContent = `Discovered Bonus Facts (${{state.discoveredFacts.size}}/{total_bonus_facts})`;
        }}
        
        if (list) {{
            list.innerHTML = '';
            state.factsList.forEach((fact, index) => {{
                const li = document.createElement('li');
                li.style.cssText = 'margin: 5px 0; padding: 5px; background: #f0f8ff; border-radius: 3px; font-size: 12px;';
                li.innerHTML = `<strong>${{index + 1}}.</strong> ${{fact}}`;
                list.appendChild(li);
            }});
        }}
    }}

    function celebrate() {{
        const state = window.virusDashboardGames[iframeId];
        // Stop timer
        if (state.timerInterval) {{
            clearInterval(state.timerInterval);
        }}
        
        const elapsed = Math.floor((Date.now() - state.startTime) / 1000);
        const minutes = Math.floor(elapsed / 60);
        const seconds = elapsed % 60;
        const timeStr = `${{minutes}}:${{seconds.toString().padStart(2, '0')}}`;
        
        // Create celebration overlay
        const celebration = document.createElement('div');
        celebration.id = `celebration-${{iframeId}}`;
        celebration.style.cssText = `
            position: absolute;
            top: 0;
            left: 0;
            width: 100%;
            height: 100%;
            background: rgba(0, 102, 204, 0.9);
            z-index: 10000;
            display: flex;
            justify-content: center;
            align-items: center;
            flex-direction: column;
            color: white;
            font-family: Arial, sans-serif;
            text-align: center;
            animation: fadeIn 0.5s ease-in;
        `;
        
        const continueButton = document.createElement('button');
        continueButton.textContent = 'Continue Exploring';
        continueButton.style.cssText = `
            background: white;
            color: #0066cc;
            border: none;
            padding: 15px 30px;
            font-size: 1.1em;
            border-radius: 25px;
            cursor: pointer;
            margin-top: 30px;
            box-shadow: 0 4px 8px rgba(0,0,0,0.2);
        `;
        continueButton.addEventListener('click', function() {{
            returnToDisplay();
        }});
        
        celebration.innerHTML = `
            <h1 style="font-size: 3em; margin: 0; animation: bounce 1s ease-in-out infinite;">🎉 Triviradae Complete! 🎉</h1>
            <h2 style="font-size: 1.5em; margin: 20px 0;">You found all {total_bonus_facts} bonus facts!</h2>
            <h3 style="font-size: 1.3em; margin: 10px 0; color: #ffff99;">⏱️ Time: ${{timeStr}}</h3>
            <p style="font-size: 1.2em; max-width: 600px; margin: 20px;">
                Excellent work exploring the virus surveillance data! You've mastered the art of trivia hunting in scientific dashboards.
            </p>
        `;
        celebration.appendChild(continueButton);
        
        // Add CSS animations if not already present
        if (!document.getElementById(`game-animations-${{iframeId}}`)) {{
            const style = document.createElement('style');
            style.id = `game-animations-${{iframeId}}`;
            style.textContent = `
                @keyframes fadeIn {{
                    from {{ opacity: 0; }}
                    to {{ opacity: 1; }}
                }}
                @keyframes bounce {{
                    0%, 20%, 50%, 80%, 100% {{ transform: translateY(0); }}
                    40% {{ transform: translateY(-10px); }}
                    60% {{ transform: translateY(-5px); }}
                }}
            `;
            document.head.appendChild(style);
        }}
        
        document.body.appendChild(celebration);
    }}

    // Initialize tracking when page loads
    document.addEventListener('DOMContentLoaded', function() {{
        // Check if we're in game mode
        const urlParams = new URLSearchParams(window.location.search);
        const isGameMode = urlParams.get('game') === 'true';
        
        if (isGameMode) {{
            // In game mode, show the game menu immediately
            createFactsDisplay();
            hideGameElements();
            setTimeout(() => {{
                showGameMenu();
            }}, 100); // Small delay to ensure page is fully loaded
        }} else {{
            // Normal mode - just create and hide game elements
            createFactsDisplay();
            hideGameElements();
        }}
    }});

}})(); // Close the IIFE (Immediately Invoked Function Expression)
</script>
    """
    
    # Insert the JavaScript before the closing body tag
    enhanced_html = html_string.replace('</body>', bonus_tracking_js + '</body>')
    
    # Write the enhanced HTML to file
    with open(html_path, 'w', encoding='utf-8') as f:
        f.write(enhanced_html)




def create_visualizations(report_data: pd.DataFrame, visibility: str = 'public',
                         test_subset: bool = False, output_dir: Optional[Path] = None,
                         generate_html: bool = True) -> Path:
    """
    Create all visualizations from report card data.
    Matches the approach from generate_report_cards.py exactly.
    """
    console = Console()
    
    if output_dir is None:
        output_dir = Path(f'outputs/visualizations_{visibility}')
    
    # Create output directories
    output_dir.mkdir(parents=True, exist_ok=True)
    (output_dir / 'by_city').mkdir(exist_ok=True)
    (output_dir / 'by_virus').mkdir(exist_ok=True)
    
    console.print(f"[bold blue]Creating {visibility} visualizations...[/bold blue]")
    
    # Prepare data for visualization (matching generate_report_cards.py approach)
    viz_data, sequencing_coverage = prepare_data_for_visualization(report_data, visibility)
    
    if viz_data.empty:
        console.print(f"[red]No data available for {visibility} visualizations[/red]")
        return output_dir
    
    # Get available cities and viruses
    available_cities = get_sites_by_visibility(visibility)
    available_viruses = get_viruses_by_visibility(visibility)
    
    # In test mode, use all cities and viruses but only save PNG files
    if test_subset:
        console.print(f"[yellow]Test mode: Generating PNG files for all {len(available_cities)} cities and {len(available_viruses)} viruses[/yellow]")
    
    # Track results
    city_results = {}
    virus_results = {}
    
    # Generate by-city visualizations
    console.print("\n[bold blue]Generating by-city visualizations...[/bold blue]")
    
    virus_categories = ['enteric', 'respiratory', 'atypical', 'animal']
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        console=console
    ) as progress:
        
        total_city_tasks = len(available_cities) * len(virus_categories)
        city_task = progress.add_task("Creating city heatmaps...", total=total_city_tasks)
        
        for city in available_cities:
            for category in virus_categories:
                try:
                    success = create_by_city_heatmap(viz_data, city, category, output_dir, visibility, sequencing_coverage, test_subset, generate_html)
                    city_results[(city, category)] = success
                except Exception as e:
                    console.print(f"[red]Error creating {category} heatmap for {city}: {e}[/red]")
                    city_results[(city, category)] = False
                
                progress.update(city_task, advance=1)
    
    # Generate by-virus visualizations
    console.print("\n[bold blue]Generating by-virus visualizations...[/bold blue]")
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        console=console
    ) as progress:
        
        virus_task = progress.add_task("Creating virus heatmaps...", total=len(available_viruses))
        
        for virus in available_viruses:
            try:
                success = create_by_virus_heatmap(viz_data, virus, output_dir, visibility, sequencing_coverage, test_subset, generate_html)
                virus_results[virus] = success
            except Exception as e:
                console.print(f"[red]Error creating heatmap for {virus}: {e}[/red]")
                virus_results[virus] = False
            
            progress.update(virus_task, advance=1)
    
    
    # Summary
    successful_cities = sum(1 for success in city_results.values() if success)
    successful_viruses = sum(virus_results.values())
    
    console.print(f"\n[bold green]✓ {visibility.title()} visualization generation complete![/bold green]")
    console.print(f"Output directory: {output_dir}")
    console.print(f"City visualizations: {successful_cities}/{len(city_results)} successful")
    console.print(f"Virus visualizations: {successful_viruses}/{len(available_viruses)} successful")
    
    return output_dir


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description='Generate virus surveillance visualizations',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # First run - retrieves and caches data
  python create_visualizations.py
  
  # Subsequent runs - uses cached data
  python create_visualizations.py --use-cache
  
  # Force fresh data retrieval
  python create_visualizations.py --no-cache
  
  # Use specific cache directory
  python create_visualizations.py --use-cache --cache-dir /tmp/virus_cache
  
  # Generate only public visualizations
  python create_visualizations.py --use-cache --visibility public
  
  # Skip HTML generation (PNG only)
  python create_visualizations.py --use-cache --no-html
        """
    )
    
    parser.add_argument('--use-cache', action='store_true',
                       help='Use cached data if available instead of querying LabKey')
    parser.add_argument('--no-cache', action='store_true',
                       help='Force fresh data retrieval (ignore existing cache)')
    parser.add_argument('--cache-dir', type=Path, default=Path(__file__).parent / 'local-outputs' / 'cache',
                       help='Directory for cache files (default: local-outputs/cache)')
    parser.add_argument('--visibility', choices=['public', 'private', 'both'], default='both',
                       help='Which visualizations to generate (default: both)')
    parser.add_argument('--output-dir', type=Path, default=None,
                       help='Output directory for visualizations (default: outputs/visualizations_<visibility>)')
    parser.add_argument('--test-filter', action='store_true',
                       help='Filter to Columbia, MO June 2024 for testing')
    parser.add_argument('--no-html', action='store_true',
                       help='Skip HTML generation, create only PNG files')
    parser.add_argument('--max-rows', type=int, default=None,
                       help='Maximum rows to retrieve from LabKey (for testing)')
    
    return parser.parse_args()


def main():
    """Main function for visualization generation with caching support."""
    console = Console()
    args = parse_args()
    
    console.print("[bold green]🔬 Virus Surveillance Visualization Generator[/bold green]\n")
    
    # Detect environment and set up output directories
    environment = detect_environment()
    output_base_dir = get_output_base_dir(environment)
    console.print(f"[cyan]Environment: {environment}[/cyan]")
    console.print(f"[cyan]Output base directory: {output_base_dir}[/cyan]")
    
    # Check cache usage
    use_cache = args.use_cache and not args.no_cache
    
    # Check if API key is available (only needed if not using cache)
    if not use_cache and not os.environ.get('LABKEY_API'):
        console.print("[yellow]⚠ LABKEY_API environment variable not set[/yellow]")
        console.print("Set your LabKey API key to retrieve data:")
        console.print("export LABKEY_API='your_api_key_here'")
        console.print("\nOr use cached data with: python create_visualizations.py --use-cache")
        return 1
    
    try:
        # Set up cache paths (use environment-aware directory if default)
        if args.cache_dir == Path(__file__).parent / 'local-outputs' / 'cache':
            cache_dir = output_base_dir / 'cache'
        else:
            cache_dir = args.cache_dir
        
        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        raw_cache_file = cache_dir / f'raw_virus_data_{timestamp}.pkl.gz'
        
        # If using cache, find the most recent cache file
        if use_cache:
            cache_files = list(cache_dir.glob('raw_virus_data_*.pkl.gz'))
            if cache_files:
                raw_cache_file = max(cache_files, key=lambda p: p.stat().st_mtime)
                console.print(f"[cyan]Using cache file: {raw_cache_file}[/cyan]")
            else:
                console.print("[yellow]No cache files found. Will retrieve fresh data.[/yellow]")
                use_cache = False
        
        # Retrieve or load raw virus data
        console.print("[blue]Getting virus data...[/blue]")
        raw_data = retrieve_all_virus_data(
            max_rows=args.max_rows,
            show_progress=True,
            test_filter=args.test_filter,
            cache_file=raw_cache_file,
            use_cache=use_cache
        )
        
        if raw_data.empty:
            console.print("[red]No raw data available[/red]")
            return 1
        
        # Create report card data
        visibility_suffix = 'test' if args.test_filter else 'all'
        report_cache_file = cache_dir / f'report_card_data_{visibility_suffix}_{timestamp}.pkl.gz'
        
        # If using cache, find matching report card cache
        if use_cache:
            report_cache_files = list(cache_dir.glob(f'report_card_data_{visibility_suffix}_*.pkl.gz'))
            if report_cache_files:
                report_cache_file = max(report_cache_files, key=lambda p: p.stat().st_mtime)
        
        console.print("[blue]Creating report card data...[/blue]")
        report_data = create_report_card_data(
            raw_data,
            test_filter=args.test_filter,
            show_progress=True,
            cache_file=report_cache_file,
            use_cache=use_cache
        )
        
        if report_data.empty:
            console.print("[red]No report card data generated[/red]")
            return 1
        
        console.print(f"[green]✓ Processed {len(report_data)} report card records[/green]")
        
        # Generate visualizations based on visibility argument
        visibilities = []
        if args.visibility in ['public', 'both']:
            visibilities.append('public')
        if args.visibility in ['private', 'both']:
            visibilities.append('private')
        
        generate_html = not args.no_html
        
        for visibility in visibilities:
            console.print(f"\n[blue]Generating {visibility.upper()} visualizations...[/blue]")
            
            output_dir = args.output_dir
            if output_dir is None:
                # Use environment-aware output directory
                output_dir = output_base_dir / f'visualizations_{visibility}'
            elif len(visibilities) > 1:
                # If generating both, create subdirectories
                output_dir = output_dir / visibility
            
            viz_output = create_visualizations(
                report_data,
                visibility=visibility,
                test_subset=False,  # Generate all visualizations
                output_dir=output_dir,
                generate_html=generate_html
            )
            
            console.print(f"[green]✓ {visibility.title()} visualizations saved to: {viz_output}[/green]")
        
        console.print("\n[green]✅ Visualization generation complete![/green]")
        
        if use_cache:
            console.print(f"\n[cyan]Cache directory: {cache_dir}[/cyan]")
            console.print("To refresh data, run without --use-cache flag")
        
        return 0
        
    except Exception as e:
        console.print(f"[red]❌ Error: {e}[/red]")
        import traceback
        traceback.print_exc()
        return 1


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)